# -*- coding: utf-8 -*-
r"""
Compressible Neo-Hookean hyperelastic material model. The tangent modulus
and the stress tensor are calculated by a user defined function.
"""
from functools import partial

import numpy as nm
from hyperelastic_ul import (filename_mesh, options, regions, fields,
    variables, ebcs, functions)

mu, K = 20., 1000.

# function which returns tangent modulus and stress tensor
def get_hyperelastic_mat(family_data, mode):
    from sfepy.terms.extmods.terms import sym2nonsym
    from sfepy.terms.terms_hyperelastic_ul import (NeoHookeanULTerm,
        BulkPenaltyULTerm)

    n_el, n_qp, sym, _= family_data.green_strain.shape
    dim = family_data.mtx_f.shape[-1]
    dim2 = dim**2

    ones = nm.ones((n_el, n_qp, 1, 1), dtype=nm.float64)
    mat_mu = ones * mu
    mat_K = ones * K

    fargs = [family_data.get(name)
             for name in NeoHookeanULTerm.family_data_names]
    stress = nm.empty((n_el, n_qp, sym, 1), dtype=nm.float64)
    NeoHookeanULTerm.stress_function(stress, mat_mu, *fargs)

    if mode == 'tan_mod':
        tanmod = nm.empty((n_el, n_qp, sym, sym), dtype=nm.float64)
        NeoHookeanULTerm.tan_mod_function(tanmod, mat_mu, *fargs)


    fargs = [family_data.get(name)
             for name in BulkPenaltyULTerm.family_data_names]
    stress_p = nm.empty((n_el, n_qp, sym, 1), dtype=nm.float64)
    BulkPenaltyULTerm.stress_function(stress_p, mat_K, *fargs)

    if mode == 'tan_mod':
        tanmod_p = nm.empty((n_el, n_qp, sym, sym), dtype=nm.float64)
        BulkPenaltyULTerm.tan_mod_function(tanmod_p, mat_K, *fargs)

        tanmod_ns = nm.zeros((n_el, n_qp, dim2, dim2), dtype=nm.float64)
        sym2nonsym(tanmod_ns, tanmod + tanmod_p)

        stress_ns = nm.zeros((n_el, n_qp, dim2, dim2), dtype=nm.float64)
        sym2nonsym(stress_ns, stress + stress_p)

        out = tanmod_ns + stress_ns

    elif mode == 'stress':
        out = stress + stress_p

    else:
        raise ValueError()

    return out


def stress_strain(out, problem, state, extend = False):
    from sfepy.base.base import Struct

    ev = partial(problem.evaluate, mode='el_avg',
                 get_hyperelastic_mat=get_hyperelastic_mat)
    stress = ev('dw_ul_he_by_fun.i.Omega(get_hyperelastic_mat, v, u)',
                term_mode='stress')
    out['cauchy_stress'] = Struct(name='output_data',
                                  mode='cell', data=stress)
    strain = ev('dw_ul_he_by_fun.i.Omega(get_hyperelastic_mat, v, u)',
                term_mode='strain')
    out['green_strain'] = Struct(name='output_data',
                                 mode='cell', data=strain)
    return out


integrals = {
    'i': 2,
}

materials = {
}

equations = {
    'balance': 'dw_ul_he_by_fun.i.Omega(get_hyperelastic_mat, v, u) = 0',
}

solvers = {
    'ls': ('ls.scipy_direct', {}),
    'newton': ('nls.newton', {
        'i_max': 20,
        'eps_a': 1e-5,
        'eps_r': 1e-3,
    }),
    'ts': ('ts.simple', {
        't0': 0,
        't1': 1,
        'n_step': 11,
    }),
}



import sys
import os
import shutil
import sysconfig
from warnings import warn

def get_top_dir():
    """
    Return SfePy installation directory information.
    """
    import os.path as op

    # If installed, up_dir is '.', otherwise (in (git) source directory) '..'.
    for up_dir in ['..', '.']:
        top_dir = op.normpath(op.realpath(op.join(op.dirname(__file__),
                                                  up_dir)))
        aux = op.join(top_dir, 'LICENSE')
        if op.isfile(aux):
            break
    else:
        print('Warning: cannot determine SfePy top level directory.')
        up_dir = top_dir = ''

    in_source_tree = up_dir == '..'

    return top_dir, in_source_tree

top_dir, in_source_tree = get_top_dir()


msg_unknown_os = """could not determine operating system!
try setting it in site_cfg.py manually, see site_cfg_template.py"""

msg_numpydoc = """could not find numpydoc!
If it is installed in a non-standard location, try setting it in
site_cfg.py manually."""

has_attr = lambda obj, attr: obj and hasattr(obj, attr)


def compose_system_compile_flags(is_posix: bool) -> list:
    """
    Provides a list of compile flags that tries to be as similar as possible
    to what Python itself was built with. This has been done historically by
    numpy.distutils (now deprecated) and a squeezed version of it is brought
    over to here.
    """
    if not is_posix:
        return []

    cflags, configure_cppflags, configure_cflags = sysconfig.get_config_vars(
        'CFLAGS', 'CONFIGURE_CPPFLAGS', 'CONFIGURE_CFLAGS')

    return (cflags + ' ' + configure_cppflags + ' ' + configure_cflags).split()


class Config(object):
    """
    Read and access the site configuration from the current directory.
    """

    def __init__(self):
        from importlib.machinery import SourceFileLoader
        from sfepy.base.base import sfepy_config_dir

        cwd = os.getcwd()
        config_filename = os.path.join(cwd, 'site_cfg.py')
        if not os.path.exists(config_filename):
            config_filename = os.path.join(sfepy_config_dir, 'site_cfg.py')

        if not os.path.exists(config_filename):
            try:
                shutil.copyfile(os.path.join(top_dir,
                                             'site_cfg_template.py'),
                                config_filename)

            except:
                pass

        try:
            self.site_cfg = (SourceFileLoader('site_cfg', config_filename)
                             .load_module())

        except FileNotFoundError:
            print(f'Warning: SfePy site configuration file ({config_filename})'
                  ' not found.')
            self.site_cfg = None

    def python_version(self):
        return "%d.%d" % tuple(sys.version_info[:2])

    def python_include(self):
        if (has_attr(self.site_cfg, 'python_include')
            and (self.site_cfg.python_include != 'auto')):
            return self.site_cfg.python_include

        else:
            return sysconfig.get_config_var('INCLUDEPY')


    def system(self):
        if (has_attr(self.site_cfg, 'system')
            and self.site_cfg.system is not None):
            return self.site_cfg.system

        else:
            if os.name in ['posix']:
                return 'posix'
            elif os.name in ['nt']:
                return 'windows'
            else:
                raise ValueError(msg_unknown_os)

    def compile_flags(self):
        if has_attr(self.site_cfg, 'compile_flags'):
            flags = self.site_cfg.compile_flags
            if isinstance(flags, str):
                warn('Compile flags should be given as a list of strings.'
                     ' Space-separated strings may be removed in the'
                     ' near future.',
                     DeprecationWarning, stacklevel=2)

                flags = flags.split()

        else:
            flags = ['-g', '-O2']

        return flags + compose_system_compile_flags(self.system() == 'posix')

    def debug_flags(self) -> list:
        if has_attr(self.site_cfg, 'debug_flags'):
            return self.site_cfg.debug_flags
        else:
            return []

    def numpydoc_path(self):
        if (has_attr(self.site_cfg, 'numpydoc_path') and
            (self.site_cfg.numpydoc_path is not None)):
            return self.site_cfg.numpydoc_path

        else:
            try:
                import numpydoc; numpydoc
            except ImportError:
                raise ValueError(msg_numpydoc)

    def is_release(self):
        if has_attr(self.site_cfg, 'is_release'):
            return self.site_cfg.is_release
        else:
            return ''

    def refmap_memory_factor(self):
        if has_attr(self.site_cfg, 'refmap_memory_factor'):
            return self.site_cfg.refmap_memory_factor
        else:
            return None

    def debug_warped_cells(self):
        if has_attr(self.site_cfg, 'debug_warped_cells'):
            return self.site_cfg.debug_warped_cells
        else:
            return False

site_config = Config()



r"""
Contact of two elastic bodies with a penalty function for enforcing the contact
constraints.

Find :math:`\ul{u}` such that:

.. math::
    \int_{\Omega} D_{ijkl}\ e_{ij}(\ul{v}) e_{kl}(\ul{u})
    + \int_{\Gamma_{c}} \varepsilon_N \langle g_N(\ul{u}) \rangle \ul{n} \ul{v}
    = 0
    \;, \quad \forall \ul{v} \;,

where :math:`\varepsilon_N \langle g_N(\ul{u}) \rangle` is the penalty
function, :math:`\varepsilon_N` is the normal penalty parameter, :math:`\langle
g_N(\ul{u}) \rangle` are the Macaulay's brackets of the gap function
:math:`g_N(\ul{u})` and

.. math::
    D_{ijkl} = \mu (\delta_{ik} \delta_{jl}+\delta_{il} \delta_{jk}) +
    \lambda \ \delta_{ij} \delta_{kl}
    \;.

This example also demonstrates use of an experimental contact term based on the
IPC Toolkit [1].

[1] https://github.com/ipc-sim/ipc-toolkit

Usage examples::

  # Check regions and EBC nodes.
  sfepy-run sfepy/examples/linear_elasticity/two_bodies_contact.py --save-regions-as-groups --save-ebc-nodes --solve-not

  sfepy-view two_bodies_ebc_nodes.vtk
  sfepy-view two_bodies_regions.h5

  # Run with default parameters and view results.
  sfepy-run sfepy/examples/linear_elasticity/two_bodies_contact.py

  sfepy-view output/contact/two_bodies.h5
  sfepy-view output/contact/two_bodies.h5 -f u:wu:f1:p0 1:vw:wu:f1:p0

  # Plot the nonlinear solver convergence.
  python3 sfepy/scripts/plot_logs.py output/contact/log.txt

  # Run with default parameters and IPC toolkit, view results.
  sfepy-run sfepy/examples/linear_elasticity/two_bodies_contact.py -d "contact='ipc'"
  sfepy-view output/contact/two_bodies.h5 -f u:wu:f1:p0 1:vw:wu:f1:p0

  # Plot the IPC contact parameters evolution.
  python3 sfepy/scripts/plot_logs.py output/contact/clog.txt
"""
import os.path as op
from functools import partial
import inspect

from sfepy.base.base import output
from sfepy.base.log import Log
from sfepy.mechanics.matcoefs import stiffness_from_youngpoisson
from sfepy.discrete.fem.meshio import UserMeshIO
from sfepy.solvers.nls import apply_line_search_ipc as apply_line_search

import numpy as nm

def get_bbox(dims, centre, eps=0.0):
    dims = nm.asarray(dims)
    centre = nm.asarray(centre)

    bbox = nm.r_[[centre - (0.5 - eps) * dims], [centre + (0.5 - eps) * dims]]
    return bbox

def gen_two_bodies(dims0, shape0, centre0, dims1, shape1, centre1, shift1):
    from sfepy.discrete.fem import Mesh
    from sfepy.mesh.mesh_generators import gen_block_mesh

    m0 = gen_block_mesh(dims0, shape0, centre0)
    m1 = gen_block_mesh(dims1, shape1, centre1)

    coors = nm.concatenate((m0.coors, m1.coors + shift1), axis=0)

    desc = m0.descs[0]
    c0 = m0.get_conn(desc)
    c1 = m1.get_conn(desc)
    conn = nm.concatenate((c0, c1 + m0.n_nod), axis=0)

    ngroups = nm.zeros(coors.shape[0], dtype=nm.int32)
    ngroups[m0.n_nod:] = 1

    mat_id = nm.zeros(conn.shape[0], dtype=nm.int32)
    mat_id[m0.n_el:] = 1

    name = 'two_bodies'

    mesh = Mesh.from_data(name, coors, ngroups, [conn], [mat_id], m0.descs)

    return mesh

def markdown_table_from_dict(adict):
    header = '| option | value |\n'
    separator = '| --- | --- |\n'
    rows = '\n'.join([f'| {key} | {val} |' for key, val in adict.items()])

    return header + separator + rows

def define(
        dims0=(1.0, 1.0, 0.5),
        shape0=(2, 2, 2),
        centre0=(0, 0, -0.25),

        dims1=(1.2, 0.8, 0.5),
        shape1=(3, 3, 2),
        centre1=(0, 0, 0.25),

        shift10=(0.0, 0.0, 1e-4),
        shift11=(0.0, 0.0, -0.1),

        young=1.0,
        poisson=0.3,
        rho=1.0,

        cm=None,
        ck=0.0,
        dhat=1e-2,
        pspd='NONE',

        t1=1000,
        n_step=5,
        contact='builtin',

        output_dir='output/contact',
        verbose=True,
):
    args = locals()

    inodir = partial(op.join, output_dir)

    output.set_output(filename=inodir('output_log.txt'), quiet=not verbose,
                      combined=True)

    signature = inspect.signature(define)
    arg_names = list(signature.parameters.keys())
    options = {name : args[name] for name in arg_names}
    table = markdown_table_from_dict(options)
    with open(inodir('options.md'), 'w') as fd:
        fd.write(table)

    dim = len(dims0)
    shape0 = shape0[:dim]
    centre0 = centre0[:dim]

    dims1 = dims1[:dim]
    shape1 = shape1[:dim]
    centre1 = centre1[:dim]

    shift10 = shift10[:dim]
    shift11 = shift11[:dim]

    shift11 = nm.array(shift11)

    clog = Log([[r'$d$'], [r'$k$'], [r'$\nabla B$'], [r'$\nabla E$']],
               xlabels=['', '', 'all iterations', 'all iterations'],
               ylabels=[r'$d$', r'$k$', r'$\nabla B$', r'$\nabla E$'],
               yscales=['log', 'linear', 'linear', 'linear'],
               is_plot=False,
               log_filename=inodir('clog.txt'),
               formats=[['%.8e']] * 4)

    def mesh_hook(mesh, mode):
        if mode == 'read':
            return gen_two_bodies(dims0, shape0, centre0,
                                  dims1, shape1, centre1, shift10)

        elif mode == 'write':
            pass

    def post_process(out, pb, state, extend=False):
        from sfepy.base.base import Struct
        from sfepy.discrete.fem import extend_cell_data

        ev = pb.evaluate
        gap = ev('dw_contact.i.Contact(contact.epss, v, u)',
                 mode='el_avg', term_mode='gap')
        gap = extend_cell_data(gap, pb.domain, 'Contact', val=0.0,
                               is_surface=True)
        out['gap'] = Struct(name='output_data',
                            mode='cell', data=gap, dofs=None)

        return out

    filename_mesh = UserMeshIO(mesh_hook)

    options = {
        'nls' : 'newton',
        'ls' : 'ls',
        'output_dir' : output_dir,
        'output_format' : 'h5',
        'post_process_hook' : 'post_process',
    }

    bbox0 = get_bbox(dims0, centre0, eps=1e-5)
    bbox1 = get_bbox(dims1, nm.asarray(centre1) + nm.asarray(shift10), eps=1e-5)

    if dim == 2:
        regions = {
            'Omega' : 'all',
            'Omega0' : 'cells of group 0',
            'Omega1' : 'cells of group 1',
            'Bottom' : ('vertices in (y < %.12e)' % bbox0[0, 1], 'facet'),
            'Top' : ('vertices in (y > %.12e)' % bbox1[1, 1], 'facet'),
            'Contact0' : ('(vertices in (y > %.12e) *v r.Omega0)' % bbox0[1, 1],
                          'facet'),
            'Contact1' : ('(vertices in (y < %.12e) *v r.Omega1)' % bbox1[0, 1],
                          'facet'),
            'Contact' : ('r.Contact0 +s r.Contact1', 'facet')
        }

    else:
        regions = {
            'Omega' : 'all',
            'Omega0' : 'cells of group 0',
            'Omega1' : 'cells of group 1',
            'Bottom' : ('vertices in (z < %.12e)' % bbox0[0, 2], 'facet'),
            'Top' : ('vertices in (z > %.12e)' % bbox1[1, 2], 'facet'),
            'Contact0' : ('(vertices in (z > %.12e) *v r.Omega0)' % bbox0[1, 2],
                          'facet'),
            'Contact1' : ('(vertices in (z < %.12e) *v r.Omega1)' % bbox1[0, 2],
                          'facet'),
            'Contact' : ('r.Contact0 +s r.Contact1', 'facet')
        }

    fields = {
        'displacement': ('real', dim, 'Omega', 1),
    }

    variables = {
        'u' : ('unknown field', 'displacement', 0),
        'v' : ('test field', 'displacement', 'u'),
    }

    ebcs = {
        'fixb' : ('Bottom', {'u.all' : 0.0}),
        'fixt' : ('Top', {'u.all' : 'move_top'}),
    }

    def move_top(ts, coors, bc, problem, **kwargs):
        val = nm.empty_like(coors)
        val[:] = ts.nt * shift11
        return val

    functions = {
        'move_top' : (move_top,),
    }

    volume = nm.prod(dims0) + nm.prod(dims1)
    mass = volume * rho

    materials = {
        'solid' : ({
            'D' : stiffness_from_youngpoisson(dim, young=young, poisson=poisson),
        },),
        'contact' : ({
            '.m' : cm if cm is not None else mass,
            '.k' : ck, # 0 = Adaptive barrier stiffness.
            '.dhat' : dhat,
            '.Pspd' : pspd,
            '.epss' : 2e+1,
        },),
    }

    integrals = {
        'i' : 2,
    }

    if contact == 'builtin':
        equations = {
            'elasticity' :
            """
               dw_lin_elastic.2.Omega(solid.D, v, u)
             + dw_contact.i.Contact(contact.epss, v, u)
             = 0
            """,
        }

    elif contact == 'ipc':
        equations = {
            'elasticity' :
            """
               dw_lin_elastic.2.Omega(solid.D, v, u)
             + dw_contact_ipc.i.Contact(
                   contact.m, contact.k, contact.dhat, contact.Pspd, v, u
               )
             = 0
            """,
        }

    else:
        raise ValueError("contact argument can be one of 'builtin', 'ipc'!")

    if contact == 'ipc':
        apply_ls = partial(apply_line_search, clog=clog)

    else:
        apply_ls = None

    solvers = {
        'ls' : ('ls.auto_direct', {}),
        'newton' : ('nls.newton', {
            'i_max' : 30,
            'eps_a' : 1e-8,
            'eps_r' : 1e-8,
            'eps_mode' : 'or',
            'macheps' : 1e-16,
            # Linear system error < (eps_a * lin_red).
            'lin_red' : None,
            'line_search_fun' : apply_ls,
            'ls_red' : 0.5,
            'ls_red_warp' : 0.5,
            'ls_on' : 1.0,
            'ls_min' : 1e-5,
            'check' : 0,
            'delta' : 1e-8,
            'log' : {'text' : inodir('log.txt')},
            'log_vlines' : 'solve',
        }),
        'ts' : ('ts.simple', {
            't0'     : 0.0,
            't1'     : t1,
            'dt'     : None,
            'n_step' : n_step,
            'quasistatic' : True,
            'verbose' : 1,
        }),
    }

    return locals()



r"""
The linear elasticity of two discs connected using multi-point constraints.

Find :math:`\ul{u}`, :math:`\ul{u_c}` such that:

.. math::
    \int_{\Omega} D_{ijkl}\ e_{ij}(\ul{v}) e_{kl}(\ul{u})
    + \sum_{i,j \in \Omega_c} K_{kl}\ ((u_c)^{(j)}_l - (u_c)^{(i)}_l)
    = 0
    \;, \quad \forall \ul{v} \;,

where

.. math::
    D_{ijkl} = \mu (\delta_{ik} \delta_{jl}+\delta_{il} \delta_{jk}) +
    \lambda \ \delta_{ij} \delta_{kl} \;,

    K_{kl} = K_{kl}(\ul{d}, \ul{k}) \;,

:math:`u_c^{(j)}` are the DOFs - two displacements and one rotation in 2D -
of the node :math:`j` in :math:`\Omega_c`, which is a subdomain composed of two
1D spring terms with the generalized stiffness matrix :math:`K_{kl}` depending
on the directions :math:`\ul{d}` of the springs and the stiffness vector
:math:`\ul{k}`.

The deformation is governed by the Dirichlet conditions applied to one of the
spring end points, see the `dofs` argument of :func:`define()` below.

Usage Examples
--------------

- Save and display boundary regions::

    sfepy-run sfepy/examples/linear_elasticity/multi_point_constraints.py --save-regions-as-groups --solve-not

    sfepy-view annulus-c_regions.vtk -2e -f Gamma1:p0 Gamma2:p1 Gamma3:p3 Gamma4:p4 --max-plots=4 --color-map=summer

- Run::

    sfepy-run sfepy/examples/linear_elasticity/multi_point_constraints.py
    sfepy-run sfepy/examples/linear_elasticity/multi_point_constraints.py -d "dofs=(0,1)"
    sfepy-run sfepy/examples/linear_elasticity/multi_point_constraints.py -d "dofs=(0,1), is_rot=False"
    sfepy-run sfepy/examples/linear_elasticity/multi_point_constraints.py -d "dofs=2"

- Display results::

    sfepy-view annulus-c.vtk -2e
    sfepy-view annulus-c.vtk -2e -f u:wu:f1:p0 1:vw:p0 u:gu:p0
"""
import numpy as nm

from sfepy.mesh.mesh_generators import gen_cylinder_mesh
from sfepy.discrete.fem.meshio import UserMeshIO
from sfepy.discrete.fem.mesh import Mesh
from sfepy.linalg import get_coors_in_ball
from sfepy.mechanics.matcoefs import stiffness_from_lame
from sfepy.mechanics.tensors import dim2sym

def get_weights(mcoors, scoors):
    n_nod, dim = scoors.shape
    out = nm.empty((n_nod, dim2sym(dim)))
    out[:] = nm.linspace(0, 10, n_nod)[:, None]
    return out

def define(dims=(1, 1, 2, 2, 0), shape=(5, 32, 0), order=1, dofs=(0, 1, 2),
           is_rot=True, output_dir='.'):
    if not isinstance(dofs, tuple):
        dofs = (dofs,)

    nuc = 2 + is_rot
    dofs0 = tuple(set(range(nuc)).difference(dofs))

    sdofs = ','.join([f'{ii}' for ii in dofs])
    sdofs0 = ','.join([f'{ii}' for ii in dofs0])

    def mesh_hook(mesh, mode):
        if mode == 'read':
            _mesh = gen_cylinder_mesh(dims, shape, (0, 0, 0),
                                     make_2d=True)
            _mesh2 = _mesh.copy()
            _mesh2.transform_coors(nm.eye(2, dtype=nm.float64) * 2.5)
            _mesh = _mesh + _mesh2

            # Add constraint vertices (and cells) to the base mesh.
            coors, vgroups, conns, mat_ids, descs = _mesh._get_io_data()
            nc = coors.shape[0]

            coors = nm.r_[coors, nm.zeros((4, 2), dtype=nm.float64)]
            vgroups = nm.r_[vgroups, [1, 2, 3, 4]]
            conns.append([[nc, nc + 1],
                          [nc + 2, nc + 3]])
            mat_ids.append([1, 2])
            descs.append('1_2')
            mesh = Mesh.from_data('annulus-c',
                                  coors, vgroups, conns, mat_ids, descs)
            return mesh

        elif mode == 'write':
            pass

    filename_mesh = UserMeshIO(mesh_hook)

    regions = {
        'Omega' : 'all',
        'Gamma1' : ('vertices by get_gamma1', 'facet'),
        'Gamma2' : ('vertices by get_gamma2', 'facet'),
        'Gamma3' : ('vertices by get_gamma3', 'facet'),
        'Gamma4' : ('vertices by get_gamma4', 'facet'),
        'Omega12C' : ('(vertices of group 1) +v (vertices of group 2)',
                      'cell', None, {'cell_tdim': 1}),
        'Gamma1C' : ('vertices of group 1', 'vertex'),
        'Gamma2C' : ('vertices of group 2', 'vertex'),
        'Omega34C' : ('(vertices of group 3) +v (vertices of group 4)',
                      'cell', None, {'cell_tdim': 1}),
        'Gamma3C' : ('vertices of group 3', 'vertex'),
        'Gamma4C' : ('vertices of group 4', 'vertex'),
        'OmegaC' : ('r.Omega12C +c r.Omega34C', 'cell', None, {'cell_tdim': 1})
    }

    centre = [0, 0]
    functions = {
        'get_gamma1' : (lambda coors, domain:
                        get_coors_in_ball(coors, centre, 1.1, 0.9),),
        'get_gamma2' : (lambda coors, domain:
                        get_coors_in_ball(coors, centre, 2.1, 1.9),),
        'get_gamma3' : (lambda coors, domain:
                        get_coors_in_ball(coors, centre, 2.6, 2.4),),
        'get_gamma4' : (lambda coors, domain:
                        get_coors_in_ball(coors, centre, 5.1, 4.9),),
        'get_weights' : (get_weights,),
    }

    fields = {
        'fu' : ('real', 2, 'Omega', order),
        'fc' : ('real', nuc, 'OmegaC', 1),
    }

    variables = {
        'u' : ('unknown field', 'fu', 0),
        'v' : ('test field',    'fu', 'u'),
        'uc' : ('unknown field', 'fc', 1),
        'vc' : ('test field',    'fc', 'uc'),
    }

    ebcs = {
        'uc' : ('Gamma1C', {f'uc.[{sdofs}]' : 0.1},),
    }
    if dofs0:
        ebcs.update({
            'uc0' : ('Gamma1C', {f'uc.[{sdofs0}]' : 0.0},),
        })
    if not is_rot:
        ebcs.update({
            'u0' : ('Gamma4', {'u.all' : 0.0}),
        })

    lcbcs = {
        'rigid' : (('Gamma1', 'Gamma2C'), {'u.all' : 'uc.all'},
                   None, 'rigid2'),
        'avg1' : (('Gamma2', 'Gamma3C'), {'u.all' : 'uc.all'},
                  'get_weights', 'average_force'),
        'avg2' : (('Gamma3', 'Gamma4C'), {'u.all' : 'uc.all'},
                  None, 'average_force'),
    }

    materials = {
        'm' : ({
            'D' : stiffness_from_lame(dim=2, lam=1e1, mu=1e0),
            'ks' : [[1e+5], [1e+5], [1e+5]][:nuc],
            'dvec' : [[0.01], [0.01]],
        },),
    }

    integrals = {
        'i' : 2 * order,
    }

    if is_rot:
        equations = {
            'eq1' :
            """dw_lin_elastic.i.Omega(m.D, v, u)
             + dw_lin_dspring_rot.0.Omega12C(m.dvec, m.ks, vc, uc)
             + dw_lin_dspring_rot.0.Omega34C(m.dvec, m.ks, vc, uc)
             = 0
            """,
        }

    else:
        equations = {
            'eq1' :
            """dw_lin_elastic.i.Omega(m.D, v, u)
             + dw_lin_dspring.0.Omega12C(m.dvec, m.ks, vc, uc)
             + dw_lin_dspring.0.Omega34C(m.dvec, m.ks, vc, uc)
             = 0
            """,
        }

    solvers = {
        'ls' : ('ls.auto_direct', {}),
        'newton' : ('nls.newton', {
            'i_max'      : 1,
            'eps_a'      : -1e-10,
            'check' : 0,
        }),
    }

    options = {
        'output_dir' : output_dir,
        'nls' : 'newton',
        'ls' : 'ls',
    }

    return locals()



"""
Flexoelectricity related terms.
"""
import numpy as nm

from sfepy.mechanics.tensors import dim2sym
from sfepy.terms.terms_multilinear import ETermBase

def make_grad2strain(dim):
    if dim == 3:
        g2s = nm.array([
           # 11 12 13 21 22 23 31 32 33
            [1, 0, 0, 0, 0, 0, 0, 0, 0], # 11
            [0, 0, 0, 0, 1, 0, 0, 0, 0], # 22
            [0, 0, 0, 0, 0, 0, 0, 0, 1], # 33
            [0, 1, 0, 1, 0, 0, 0, 0, 0], # 12
            [0, 0, 1, 0, 0, 0, 1, 0, 0], # 13
            [0, 0, 0, 0, 0, 1, 0, 1, 0], # 23
        ], dtype=nm.float64)
        g2s = g2s.reshape((1, 1, 6, 9))

    elif dim == 2:
        g2s = nm.array([
           # 11 12 21 22
            [1, 0, 0, 0], # 11
            [0, 0, 0, 1], # 22
            [0, 1, 1, 0], # 12
        ], dtype=nm.float64)
        g2s = g2s.reshape((1, 1, 3, 4))

    elif dim == 1:
        g2s = nm.array([
            [1],
        ], dtype=nm.float64)
        g2s = g2s.reshape((1, 1, 1, 1))

    else:
        raise ValueError(f'space dimension must be 1, 2, or 3! (is {dim})')

    return g2s

class MixedStrainGradElasticTerm(ETermBase):
    r"""
    Flexoelectric strain gradient elasticity term, mixed formulation.

    Additional evaluation modes:

      - `'strain'` - compute strain from the displacement gradient (state)
        variable.

    :Definition:

    .. math::
        \int_{\Omega} a_{ijklmn}\ e_{ij,k}(\ull{\delta w}) \ e_{lm,n}(\ull{w})

    :Arguments:
        - material: :math:`a_{ijklmn}`
        - virtual/parameter_1: :math:`\ull{\delta w}`
        - state/parameter_2: :math:`\ull{w}`
    """
    name = 'de_m_sg_elastic'
    arg_types = (('material', 'virtual', 'state'),
                 ('material', 'parameter_1', 'parameter_2'))
    arg_shapes = {'material' : 'SD, SD', 'virtual' : ('D2', 'state'),
                  'state' : 'D2', 'parameter_1' : 'D2', 'parameter_2' : 'D2'}
    modes = ('weak', 'eval')

    def get_function(self, mat, virtual, state, mode=None, term_mode=None,
                     diff_var=None, **kwargs):
        dim = self.region.dim
        sym = dim2sym(dim)
        aux = make_grad2strain(dim)
        n_qp = mat.shape[1]
        mat = mat.reshape((-1, n_qp, dim, sym, dim, sym))

        if term_mode is None:
            return self.make_function(
                'kIlJ,Ii,Jj,i.k,j.l',
                (mat, 'mat'), (aux, 'aux1'), (aux, 'aux2'), virtual, state,
                mode=mode, diff_var=diff_var,
            )

        elif term_mode == 'strain':
            return self.make_function(
                'Ii,i', (aux, 'aux1'), state, mode=mode, diff_var=None,
            )

        elif term_mode == 'grad_strain':
            return self.make_function(
                'Jj,j.l', (aux, 'aux2'), state, mode=mode, diff_var=None,
            )

        elif term_mode == 'double_stress':
            return self.make_function(
                'kIlJ,Ii,Jj,j.l',
                (mat, 'mat'), (aux, 'aux1'), (aux, 'aux2'), state,
                mode=mode, diff_var=None,
            )

        else:
            raise ValueError('unsupported term mode in %s! (%s)'
                             % (self.name, term_mode))

class MixedFlexoCouplingTerm(ETermBase):
    r"""
    Flexoelectric coupling term, mixed formulation.

    :Definition:

    .. math::
        \int_{\Omega} f_{ijkl}\ e_{jk,l}(\ull{\delta w}) \nabla_i p \\
        \int_{\Omega} f_{ijkl}\ e_{jk,l}(\ull{w}) \nabla_i q

    :Arguments 1:
        - material: :math:`f_{ijkl}`
        - virtual/parameter_t: :math:`\ull{\delta w}`
        - state/parameter_s: :math:`p`

    :Arguments 2:
        - material: :math:`f_{ijkl}`
        - state    : :math:`\ull{w}`
        - virtual  : :math:`q`
    """
    name = 'de_m_flexo_coupling'
    arg_types = (('material', 'virtual', 'state'),
                 ('material', 'state', 'virtual'),
                 ('material', 'parameter_t', 'parameter_s'))
    arg_shapes = [{'material' : 'D, SD',
                   'virtual/dw-p' : ('D2', None), 'state/dw-p' : 1,
                   'virtual/dp-w' : (1, None), 'state/dp-w' : 'D2',
                   'parameter_t' : 'D2', 'parameter_s' : 1}]
    modes = ('dw-p', 'dp-w', 'eval')

    def get_function(self, mat, tvar, svar, mode=None, term_mode=None,
                     diff_var=None, **kwargs):
        dim = self.region.dim
        sym = dim2sym(dim)
        aux = make_grad2strain(dim)
        n_qp = mat.shape[1]
        mat = mat.reshape((-1, n_qp, dim, dim, sym))

        if term_mode is None:
            fun = self.make_function(
                'jkI,Ii,i.k,0.j',
                (mat, 'mat'), (aux, 'aux'), tvar, svar,
                mode=mode, diff_var=diff_var,
            )

        elif term_mode == 'electric_displacement':
            fun = self.make_function(
                'jkI,Ii,i.k',
                (mat, 'mat'), (aux, 'aux'), tvar,
                mode=mode, diff_var=diff_var,
            )

        elif term_mode == 'double_stress':
            fun = self.make_function(
                'jkI,Ii,0.j',
                (mat, 'mat'), (aux, 'aux'), svar,
                mode=mode, diff_var=diff_var,
            )

        else:
            raise ValueError('unsupported term mode in %s! (%s)'
                             % (self.name, term_mode))

        return fun

class MixedFlexoTerm(ETermBase):
    r"""
    Mixed formulation displacement gradient consistency term.

    :Definition:

    .. math::
        \int_{\Omega} v_{i,j} a_{ij} \\
        \int_{\Omega} u_{i,j} \delta a_{ij}

    :Arguments 1:
        - virtual/parameter_v: :math:`\ul{v}`
        - state/parameter_t: :math:`\ull{a}`

    :Arguments 2:
        - state    : :math:`\ul{u}`
        - virtual  : :math:`\ull{\delta a}`
    """
    name = 'de_m_flexo'
    arg_types = (('virtual', 'state'),
                 ('state', 'virtual'),
                 ('parameter_v', 'parameter_t'))
    arg_shapes = [{'virtual/du-a' : ('D', None), 'state/du-a' : 'D2',
                   'virtual/da-u' : ('D2', None), 'state/da-u' : 'D',
                   'parameter_v' : 'D', 'parameter_t' : 'D2'},
                  {'opt_material' : None}]
    modes = ('du-a', 'da-u', 'eval')

    def get_function(self, vvar, tvar, mode=None, term_mode=None,
                     diff_var=None, **kwargs):
        fun = self.make_function(
            'v(i.j)->I,I', vvar, tvar, mode=mode, diff_var=diff_var,
        )

        return fun



# SfePy version
__version__ = '2025.3'

# "Minimal" supported versions.
NUMPY_MIN_VERSION = '1.3'
SCIPY_MIN_VERSION = '0.7'
MATPLOTLIB_MIN_VERSION = '0.99.0'
PYPARSING_MIN_VERSION = '1.5.0'
PYTABLES_MIN_VERSION = '2.1.2'
MAYAVI_MIN_VERSION = '3.3.0'
SYMPY_MIN_VERSION = '0.7.3'
IGAKIT_MIN_VERSION = '0.1'
PETSC4PY_MIN_VERSION = '3.4'
SLEPC4PY_MIN_VERSION = '3.4'
MPI4PY_MIN_VERSION = '1.3.1'
PYMETIS_MIN_VERSION = '2014.1'
SCIKIT_UMFPACK_MIN_VERSION = '0.1'
MESHIO_MIN_VERSION='4.0.0'
PSUTIL_MIN_VERSION='1.0.0'
PYVISTA_MIN_VERSION='0.23.0'
OPT_EINSUM_MIN_VERSION='3.0.0'
JAX_MIN_VERSION='0.2.0'
DASK_MIN_VERSION='2.0.0'
PRIMME_MIN_VERSION='3.0.0'
OCT2PY_MIN_VERSION='5.6.0'
PYTHON_MUMPS_MIN_VERSION='0.0.3'
MUMPSPY_MIN_VERSION='1.2.0'
IPCTK_MIN_VERSION='1.3.0'
PYPARDISO_MIN_VERSION='0.4.6'

CYTHON_MIN_VERSION = '0.14.1'

def get_version(version=__version__):
    """
    Get sfepy version.

    Appends current git commit hash to `version` if `is_release` is set to True
    in the site configuration.
    """
    import os.path as op
    from sfepy.config import top_dir, site_config

    if not site_config.is_release():
        # Append current git commit hash to __version__.
        master = op.join(top_dir, '.git/refs/heads/master')
        if op.isfile(master):
            fd = open(master, 'r')
            version += '+git.%s' % fd.readline().strip()
            fd.close()

    return version

__version__ = get_version(__version__)



# -*- coding: utf-8 -*-
r"""
Nearly incompressible hyperelastic material model with active fibres.

Large deformation is described using the total Lagrangian formulation.
Models of this kind can be used in biomechanics to model biological
tissues, e.g. muscles.

Find :math:`\ul{u}` such that:

.. math::
    \intl{\Omega\suz}{} \left( \ull{S}\eff(\ul{u})
    + K(J-1)\; J \ull{C}^{-1} \right) : \delta \ull{E}(\ul{v}) \difd{V}
    = 0
    \;, \quad \forall \ul{v} \;,

where

.. list-table::
   :widths: 20 80

   * - :math:`\ull{F}`
     - deformation gradient :math:`F_{ij} = \pdiff{x_i}{X_j}`
   * - :math:`J`
     - :math:`\det(F)`
   * - :math:`\ull{C}`
     -  right Cauchy-Green deformation tensor :math:`C = F^T F`
   * - :math:`\ull{E}(\ul{u})`
     - Green strain tensor :math:`E_{ij} = \frac{1}{2}(\pdiff{u_i}{X_j} +
       \pdiff{u_j}{X_i} + \pdiff{u_m}{X_i}\pdiff{u_m}{X_j})`
   * - :math:`\ull{S}\eff(\ul{u})`
     - effective second Piola-Kirchhoff stress tensor

The effective stress :math:`\ull{S}\eff(\ul{u})` incorporates also the
effects of the active fibres in two preferential directions:

.. math::
    \ull{S}\eff(\ul{u}) = \mu J^{-\frac{2}{3}}(\ull{I}
    - \frac{1}{3}\tr(\ull{C}) \ull{C}^{-1})
    + \sum_{k=1}^2 \tau^k \ull{\omega}^k
    \;.

The first term is the neo-Hookean term and the sum add contributions of
the two fibre systems. The tensors :math:`\ull{\omega}^k =
\ul{d}^k\ul{d}^k` are defined by the fibre system direction vectors
:math:`\ul{d}^k` (unit).

For the one-dimensional tensions :math:`\tau^k` holds simply (:math:`^k`
omitted):

.. math::
    \tau = A f_{\rm max} \exp{\left\{-(\frac{\epsilon - \varepsilon_{\rm
    opt}}{s})^2\right\}} \mbox{ , } \epsilon = \ull{E} : \ull{\omega}
    \;.

Usage Examples
--------------

- Run with the Newton solver::

    sfepy-run sfepy/examples/large_deformation/active_fibres.py

- Run with the matrix-free Newton-Krylov solver from SciPy::

    sfepy-run sfepy/examples/large_deformation/active_fibres.py -d solver=root

- Visualize the Green strain tensor magnitude on the deforming mesh::

    sfepy-view output/hsphere8.h5 -f green_strain:wu:f1:p0 1:vw:o0.3:p0 --color-limits=0.0,1.3

- Visualize the stresses in active fibres on the deforming mesh::

    sfepy-view output/hsphere8.h5 -f f1_stress:wu:f1:p0 1:vw:o0.3:p0 --color-limits=0,20
    sfepy-view output/hsphere8.h5 -f f2_stress:wu:f1:p0 1:vw:o0.3:p0 --color-limits=0,30

    sfepy-view output/hsphere8.h5 -f f1_stress:wu:f1:p0 1:vw:o0.3:p0 f2_stress:wu:f1:p1 1:vw:o0.3:p1 --grid-vector1="1.2,-1.2,0" --color-limits=0,30
"""
from functools import partial

import numpy as nm

from sfepy import data_dir
from sfepy.base.base import Struct
from sfepy.base.ioutils import edit_filename
from sfepy.homogenization.utils import define_box_regions

def get_pars_fibres(ts, coors, mode=None, which=0, vf=1.0, **kwargs):
    """
    Parameters
    ----------
    ts : TimeStepper
        Time stepping info.
    coors : array_like
        The physical domain coordinates where the parameters shound be defined.
    mode : 'qp' or 'special'
        Call mode.
    which : int
        Fibre system id.
    vf : float
        Fibre system volume fraction.
    """
    if mode != 'qp': return

    fmax = 100.0
    eps_opt = 0.01
    s = 1.0

    tt = ts.time * 2.0 * nm.pi

    x, y, z = coors[:,0], coors[:,1], coors[:,2]
    # Spherical coordinates.
    r = nm.sqrt(x**2 + y**2 + z**2)
    theta = nm.arccos(z / r)[:,None]
    phi = nm.arctan2(y, x)[:,None]
    act = 0.5 * (1.0 + nm.sin(tt - (0.5 * nm.pi)))

    if which == 0: # system 1 - meridians
        fdir = nm.c_[nm.cos(theta) * nm.cos(phi),
                     nm.cos(theta) * nm.sin(phi),
                     - nm.sin(theta)]

    elif which == 1: # system 2 - parallels
        fdir = nm.c_[-nm.sin(phi), nm.cos(phi), nm.zeros_like(phi)]

    else:
        raise ValueError(f'unknown fibre system {which}!')

    nfdir = nm.linalg.norm(fdir, axis=1, keepdims=True)
    fdir /= nm.where(nfdir > 0.0, nfdir, 1.0)

    shape = (coors.shape[0], 1, 1)
    out = {
        'fmax' : vf * nm.tile(fmax, shape),
        'eps_opt' : nm.tile(eps_opt, shape),
        's' : nm.tile(s, shape),
        'fdir' : fdir.reshape((coors.shape[0], 3, 1)),
        'act' : nm.tile(act, shape),
    }

    return out

def stress_strain(out, pb, state, extend=False):
    ev = partial(pb.evaluate, mode='el_avg', verbose=False)

    strain = ev('dw_tl_he_neohook.i.Omega(solid.mu, v, u)', term_mode='strain')
    out['green_strain'] = Struct(name='result', mode='cell', data=strain)

    stress = ev('dw_tl_he_neohook.i.Omega(solid.mu, v, u)', term_mode='stress')
    out['neohook_stress'] = Struct(name='result', mode='cell', data=stress)

    stress = ev('dw_tl_bulk_penalty.i.Omega(solid.K, v, u)', term_mode= 'stress')
    out['bulk_stress'] = Struct(name='output_data', mode='cell', data=stress)

    stress = ev(
        'dw_tl_fib_a.i.Omega(f1.fmax, f1.eps_opt, f1.s, f1.fdir, f1.act, v, u)',
        term_mode= 'stress',
    )
    out['f1_stress'] = Struct(name='result', mode='cell', data=stress)

    stress = ev(
        'dw_tl_fib_a.i.Omega(f2.fmax, f2.eps_opt, f2.s, f2.fdir, f2.act, v, u)',
        term_mode= 'stress',
    )
    out['f2_stress'] = Struct(name='result', mode='cell', data=stress)

    ts = pb.get_timestepper()
    if ts.step == 0:
        coors = pb.domain.get_mesh_coors()
        fout = {}
        for ii in range(2):
            aux = get_pars_fibres(ts, coors, mode='qp', which=ii, vf=1.0)
            fout[f'fdir{ii}'] = Struct(name='result', mode='vertex',
                                       data=aux['fdir'])

        filename = edit_filename(pb.get_output_name(),
                                 suffix='_fibres', new_ext='.vtk')
        pb.save_state(filename, out=fout)

    return out

def define(solver='newton', refine=0, n_step=21, output_dir='output'):
    filename_mesh = data_dir + '/meshes/3d/hsphere8.vtk'

    vf_matrix = 0.5
    vf_fibres1 = 0.2
    vf_fibres2 = 0.3

    options = {
        'nls' : solver,
        'ls' : 'ls',
        'ts' : 'ts',
        'refinement_level' : refine,
        'save_times' : 'all',
        'post_process_hook' : 'stress_strain',
        'output_format' : 'h5',
        'output_dir' : output_dir,
    }

    fields = {
        'displacement': (nm.float64, 3, 'Omega', 1),
    }

    materials = {
        'solid' : ({
            'K'  : vf_matrix * 1e3, # bulk modulus
            'mu' : vf_matrix * 20e0, # shear modulus of neoHookean term
        },),
        'f1' : 'get_pars_fibres1',
        'f2' : 'get_pars_fibres2',
    }

    functions = {
        'get_pars_fibres1' : (lambda ts, coors, mode=None, **kwargs:
                              get_pars_fibres(ts, coors, mode=mode, which=0,
                                              vf=vf_fibres1, **kwargs),),
        'get_pars_fibres2' : (lambda ts, coors, mode=None, **kwargs:
                              get_pars_fibres(ts, coors, mode=mode, which=1,
                                              vf=vf_fibres2, **kwargs),),
    }

    variables = {
        'u' : ('unknown field', 'displacement', 0),
        'v' : ('test field', 'displacement', 'u'),
    }

    dim = 3
    bbox = [[0] * dim, [25e-6] * dim]
    bregions = define_box_regions(dim, bbox[0], bbox[1])
    regions = {
        'Omega' : 'all',
        'Left' : bregions['Left'],
        'Near' : bregions['Near'],
        'Bottom' : bregions['Bottom'],
    }

    ebcs = {
        'e0' : ('Left', {'u.0' : 0.0}),
        'e1' : ('Near', {'u.1' : 0.0}),
        'e2' : ('Bottom', {'u.2' : 0.0}),
    }

    integrals = {
        'i' : 2,
    }
    equations = {
        'balance' : """
            dw_tl_he_neohook.i.Omega(solid.mu, v, u)
          + dw_tl_bulk_penalty.i.Omega(solid.K, v, u)
          + dw_tl_fib_a.i.Omega(f1.fmax, f1.eps_opt, f1.s, f1.fdir, f1.act, v, u)
          + dw_tl_fib_a.i.Omega(f2.fmax, f2.eps_opt, f2.s, f2.fdir, f2.act, v, u)
          = 0
        """,
    }

    solvers = {
        'ls' : ('ls.auto_direct', {}),
        'newton' : ('nls.newton', {
            'i_max'      : 10,
            'eps_a'      : 0.0,
            'eps_r'      : 1e-5,
            'eps_mode'   : 'or',
            'macheps'    : 1e-16,
            'lin_red'    : None, # Linear system error < (eps_a * lin_red).
            'ls_red'     : 0.1,
            'ls_red_warp': 0.001,
            'ls_on'      : 10.0,
            'ls_min'     : 1e-5,
            'check'      : 0,
            'delta'      : 1e-6,
            'report_status' : True,
        }),
        'root' : ('nls.scipy_root', {
            'method' : 'krylov',
            'use_jacobian' : False,
            'report_status' : True,
            'options' : {
                'ftol' : 1e-3,
            },
        }),
        'ts' : ('ts.simple', {
            't0' : 0.0,
            't1' : 1.0 * n_step / 21.0,
            'dt' : None,
            'quasistatic' : True,
            'n_step' : n_step, # has precedence over dt.
            'verbose' : 1,
        }),
    }

    return locals()
